/*
 *  SokoBan, a Sokoban game.
 *  Copyright (C) 2008  Colomban Wendling <ban@herbesfolles.org>
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


/* FIXME:
 * don't use images for angles, prefer a function to draw a quart of circle.
 */


#include <string.h>
#include <ctype.h>

#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include <SDL/SDL_ttf.h>

#include "MYSDL_CopySurface.h"

#include "sokoban.h"
#include "object.h"
#include "i18n.h"
#include "dialog.h"


static int BORDERX = 0;
static int BORDERY = -10;

static   Object *_topleft = NULL;
static   Object *_topright = NULL;
static   Object *_bottomleft = NULL;
static   Object *_bottomright = NULL;
static   Object *_background = NULL;
static TTF_Font *_font = NULL;
static TTF_Font *_fontSmall = NULL;
static TTF_Font *_fontMedium = NULL;
static TTF_Font *_fontBig = NULL;
static MyBool _initSuccess = FALSE;


MyBool dialog_init (void)
{
   trace ();
   
   #define ___LOADDATA(dst, img) \
      dst = object_newWithData (IMG_Load (DATADIR "images/" img), 0, 0)
   
   ___LOADDATA (_topleft, "tl.png");
   ___LOADDATA (_topright, "tr.png");
   ___LOADDATA (_bottomleft, "bl.png");
   ___LOADDATA (_bottomright, "br.png");
   ___LOADDATA (_background, "bg.png");
   
   #undef ___LOADDATA
   
   #define ___LOADFONT(dst, fnt, size) \
      dst = TTF_OpenFont (DATADIR fnt, size); \
      if (!dst) \
      { \
         fprintf (stderr, _("** Failed to load font: %s\n"), TTF_GetError ()); \
         dialog_exit (); \
         return FALSE; \
      }
   
   ___LOADFONT (_fontSmall, "sans.ttf", SMALL);
   ___LOADFONT (_fontMedium, "sans.ttf", MEDIUM);
   ___LOADFONT (_fontBig, "sans.ttf", BIG);
   
   #undef ___LOADFONT
   
   dialog_setFontSize (BIG);
   
   _initSuccess = TRUE;
   return _initSuccess;
}

void dialog_exit (void)
{
   trace ();
   
   object_free (_topleft), _topleft = NULL;
   object_free (_topright), _topright = NULL;
   object_free (_bottomleft), _bottomleft = NULL;
   object_free (_bottomright), _bottomright = NULL;
   object_free (_background), _background = NULL;
   
   TTF_CloseFont (_fontSmall), _fontSmall = NULL;
   TTF_CloseFont (_fontMedium), _fontMedium = NULL;
   TTF_CloseFont (_fontBig), _fontBig = NULL;
   _font = NULL;
   
   _initSuccess = FALSE;
}

static void _dialog_drawBackground (SDL_Surface *screen,
                                    int w, int h, int x, int y)
{
   int i, j;
   
   object_setPos (_topleft,
      ((x/2) - (w/2) - object_getSurfaceWidth (_topleft) - BORDERX),
      ((y/2) - (h/2) - object_getSurfaceHeight (_topleft) - BORDERY));
   object_setPos (_topright,
      ((x/2) + (w/2) + BORDERX),
      ((y/2) - (h/2) - object_getSurfaceHeight (_topright) - BORDERY));
   object_setPos (_bottomleft,
      ((x/2) - (w/2) - object_getSurfaceWidth (_bottomleft) - BORDERX),
      ((y/2) + (h/2) + BORDERY));
   object_setPos (_bottomright,
      ((x/2) + (w/2) + BORDERX),
      ((y/2) + (h/2) + BORDERY));
   
   object_blit (_topleft, screen);
   object_blit (_topright, screen);
   object_blit (_bottomleft, screen);
   object_blit (_bottomright, screen);
   
   
   /* fond entre le coin haut gauche et haut droit */
   for (i=0; i < object_getSurfaceHeight (_topleft); i++)
   {
      for (j=0; j < (w + (BORDERX * 2)); j++)
      {
         object_setPos (_background,
            (j + (x/2) - (w/2) - BORDERX),
            (i + (y/2) - (h/2) - object_getSurfaceHeight (_topleft) - 
                                                                     BORDERY));
         
         object_blit (_background, screen);
      }
   }
   /* fond du milieu */
   for (i=0; i < (h + (BORDERY * 2)); i++)
   {
      for (j=0;
           j < (w + (object_getSurfaceWidth (_topleft) * 2) + (BORDERX * 2));
           j++)
      {
         object_setPos (_background,
            (j + (x/2) - (w/2) - object_getSurfaceWidth (_topleft) -
                                                                      BORDERX),
            (i + (y/2) - (h/2) - BORDERY));
         
         object_blit (_background, screen);
      }
   }
   /* fond du milieu */
   for (i=0; i < object_getSurfaceHeight (_bottomleft); i++)
   {
      for (j=0; j < (w + (BORDERX * 2)); j++)
      {
         object_setPos (_background,
            (j + (x/2) - (w/2) - BORDERX),
            (i + (y/2) + (h/2) + BORDERY));
         
         object_blit (_background, screen);
      }
   }
}


void dialog_setFontSize (FontSize size)
{
   trace ();
   
   switch (size)
   {
      case SMALL:
         _font = _fontSmall; break;
      
      case MEDIUM:
         _font = _fontMedium; break;
      
      case BIG:
         _font = _fontBig; break;
      
      default:
         fprintf (stderr, _("** Intenal error: incorrect font size\n"));
   }
   
   BORDERX = size;
}

/**
 * \brief draw a dialog message
 * \param screen The SDL_Surface shere blit message
 * \param x The position in X where blit message (see below for details)
 * \param y The position in Y where blit message (see below for details)
 * \param str the message to draw
 * 
 * This function draw a message dialog of the \p str string at the center of
 * the position described by the rectangle starting at {0, 0} and ending at
 * {\p x, \p y} to the \p screen SDL_Surface.
 */
void dialog_info (SDL_Surface *screen, int x, int y, const char* str)
{
   trace ();
   
   SDL_Color textColor = {255, 255, 255, 255};
   SDL_Surface *ttf_text = NULL;
   Object *text;
   int textWidth, textHeight;
   
   
   if (!_initSuccess)
      return;
   
   if (!(text = object_new ()))
      return;
   
   /* this test is to be able to render string with 0 length (fake) */
   if (str[0] != '\0')
      ttf_text = TTF_RenderUTF8_Blended (_font, str, textColor);
   else
      ttf_text = TTF_RenderUTF8_Blended (_font, " ", textColor);
   
   if (!ttf_text)
   {
      object_free (text);
      return;
   }
   
   object_setSurface (text, ttf_text);
   /* arondi des hauteurs & largeurs */
   textWidth = ((object_getSurfaceWidth (text) >> 1) << 1);// + 2;
   textHeight = ((object_getSurfaceHeight (text) >> 1) << 1);// + 2;
   
   object_setPos (text, (x/2) - (textWidth/2), (y/2) - (textHeight/2));
   
   _dialog_drawBackground (screen, textWidth, textHeight, x, y);
   object_blit (text, screen);
   
   //SDL_Flip (screen);
   
   object_free (text);
}


/* Bad hack to make possible multiline text draw */
void dialog_multiline (SDL_Surface *screen, int x, int y, const char* str)
{
   trace ();
   
   SDL_Color textColor = {255, 255, 255, 255};
   Object *text;
   int i, j;
   int n = 0;
   int textWidth = 0, textHeight = 0;
   char *strp;
   SDL_Surface *s;
   
   if (!_initSuccess)
      return;
   
   strp = malloc ((strlen (str) + 1) * sizeof *strp);
   if (!strp)
   {
      memalloc_errormsg ();
      return;
   }
   
   /* first pass: get height of combined lines */
   for (i=0, j=0; str[i] != '\0'; i++)
   {
      /* explode line at \n */
      if (str[i] == '\n')
      {
         strp[j] = '\0';
         
         //printf (" > %s\n", strp); // debugging
         
         
         
         s = TTF_RenderUTF8_Solid (_font, strp, textColor);
         if (!s)
         {
            fprintf (stderr, _("** TTF error: %s\n"), TTF_GetError ());
            
            free (strp);
            
            return;
         }
         
         if (textWidth < s->w)
            textWidth = ((s->w >> 1) << 1);// +2;
         textHeight += ((s->h >> 1) << 1);// +2;
         
         SDL_FreeSurface (s); /* free temporary surface */
         
         
         
         j = 0;
      }
      else
      {
         strp[j] = str[i];
         j++;
      }
   }
   
   /* draw background of all lines */
   _dialog_drawBackground (screen, textWidth, textHeight, x, y);
   
   /* second pass: draw text above the background */
   for (i=0, j=0; str[i] != '\0'; i++)
   {
      /* explode line at \n */
      if (str[i] == '\n')
      {
         strp[j] = '\0';
         
         //printf (" > %s\n", strp); // debugging
         
         
         
         if (!(text = object_new ()))
            break;
         
         object_setSurface (text,
                            TTF_RenderUTF8_Blended (_font, strp, textColor));
         
         object_setPos (text, (x/2) - (textWidth/2), (y/2) -
            (textHeight/2) + (object_getSurfaceHeight (text)*n));
         
         object_blit (text, screen);
         
         object_free (text); /* free blitted text */
         
         
         j = 0;
         n++;
      }
      else
      {
         strp[j] = str[i];
         j++;
      }
   }
   
   free (strp);
}

/**
 * \brief return character from SDLKey or EOF if caracter is not valid
 */
static char _dialog_isValidChar (SDLKey k)
{
   if (!(KMOD_SHIFT & SDL_GetModState ()))
   {
      if (isgraph (k) || k == ' ')
         return k;
   }
   else
   {
      switch (k)
      {
         case 'a' :
         case 'z' :
         case 'e' :
         case 'r' :
         case 't' :
         case 'y' :
         case 'u' :
         case 'i' :
         case 'o' :
         case 'p' :
         case 'q' :
         case 's' :
         case 'd' :
         case 'f' :
         case 'g' :
         case 'h' :
         case 'j' :
         case 'k' :
         case 'l' :
         case 'm' :
         case 'w' :
         case 'x' :
         case 'c' :
         case 'v' :
         case 'b' :
         case 'n' : return toupper (k);
         
         case '<' : return '>';
         
         case ',' : return '?';
         case ';' : return '.';
         case ':' : return '/';
         //case '!' : return '§';
         
         case '&' : return '1';
         //case 'é' : return '2';
         case '"' : return '3';
         case '\'' : return '4';
         case '(' : return '5';
         case '-' : return '6';
         //case 'è' : return '7';
         case '_' : return '8';
         //case 'ç' : return '9';
         //case 'à' : return '0';
         //case ')' : return '°';
         case '=' : return '+';
         
         default: break;
      }
   }
   
   /* keypad */
   switch (k)
   {
      case SDLK_KP0: return '0';
      case SDLK_KP1: return '1';
      case SDLK_KP2: return '2';
      case SDLK_KP3: return '3';
      case SDLK_KP4: return '4';
      case SDLK_KP5: return '5';
      case SDLK_KP6: return '6';
      case SDLK_KP7: return '7';
      case SDLK_KP8: return '8';
      case SDLK_KP9: return '9';
      case SDLK_KP_PERIOD: return '.';
      case SDLK_KP_DIVIDE: return '/';
      case SDLK_KP_MULTIPLY: return '*';
      case SDLK_KP_MINUS: return '-';
      case SDLK_KP_PLUS: return '+';
      case SDLK_KP_EQUALS: return '=';
      
      default: break;
   }
   
   return EOF;
}


char* dialog_inputWithDefaultText (SDL_Surface *screen,
                                   const char* text, int x, int y)
{
   char* input = malloc (INPUT_MAXLEN);
   SDL_Event ev;
   MyBool loop = TRUE;
   unsigned int i = 0;
   
   /* background */
   SDL_Rect bgpos;
   SDL_Surface * bg = MYSDL_CopySurface (screen);
   bgpos.x = bgpos.y = 0;
   
   if (!bg) fprintf (stderr, _("** Internal error: background copy failed\n"));
   
   
   if (input == NULL)
   {
      memalloc_errormsg ();
      return NULL;
   }
   
   /* init display */
   if (bg) SDL_BlitSurface (bg, NULL, screen, &bgpos);
   if (text)
   {
      for (; i < INPUT_MAXLEN && text[i] != '\0'; i++)
         input[i] = text[i];
      
      input[i] = '\0';
   }
   else
      input[0] = '\0';
   
   dialog_info (screen, x, y, input);
   SDL_Flip (screen);
      
      
   #define ___aboart() \
      free (input); input = NULL; \
      loop = FALSE; continue;
   
   
   /* event loop to get user actions... */
   while (SDL_WaitEvent (&ev) && loop)
   {
      switch (ev.type)
      {
         case SDL_QUIT:
            ___aboart ();
            break;
         
         case SDL_KEYDOWN:
            switch (ev.key.keysym.sym)
            {
               case SDLK_RETURN:
               case SDLK_KP_ENTER:
                  loop = FALSE;
                  continue;
                  break;
               
               case SDLK_ESCAPE:
                  ___aboart ();
                  break;
               
               case SDLK_BACKSPACE:
                  if (i > 0)
                  {
                     i --;
                     
                     input[i] = '\0';
                  }
                  break;
               
               default:
                  {
                     char c;
                     
                     if ((c = _dialog_isValidChar (ev.key.keysym.sym)) != EOF)
                     {
                        if (i+1 < INPUT_MAXLEN)
                        {
                           input[i] = c;
                           input[i +1] = '\0';
                           
                           i++;
                        }
                        else
                           fprintf (stderr, _("** input string too long\n"));
                     }
                     else
                     {
                        #ifdef DEBUG
                        fprintf (stderr, "** invalid character\n");
                        #endif
                        continue;
                     }
                  }
            }
            break;
         
         default:
            continue;
      }
      
      #ifdef DEBUG
      printf ("string: %s\n", input);
      #endif
      
      if (bg) SDL_BlitSurface (bg, NULL, screen, &bgpos);
      dialog_info (screen, x, y, input);
      SDL_Flip (screen);
   }
   
   #undef aboart
   
   if (bg) SDL_FreeSurface (bg);
   
   return input;
}
