/*
 *  SokoBan, a Sokoban game.
 *  Copyright (C) 2008  Colomban Wendling <ban@herbesfolles.org>
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <string.h>

#include <SDL/SDL.h>

#include "map.h"
#include "dialog.h"
#include "player.h"
#include "gallery.h"
#include "level.h"
#include "set.h"
#include "i18n.h"
#include "editor.h"


#define editor_playerRemove() player_setPos (0, 0)
#define editor_playerSetPos(x,y) player_setPos (x, y)


static Object *_mouse = NULL;
static SDL_Rect _mousePos;


#define editor_setMousePosX(px) (_mousePos.x = px)
#define editor_setMousePosY(py) (_mousePos.y = py)
#define editor_setMousePos(px,py) \
   (editor_setMousePosX (px), editor_setMousePosY (py))


static void editor_helpRender (SDL_Surface *screen)
{
   dialog_setFontSize (BIG);
   dialog_info (screen, screen->w, screen->h*0.25, _("Editor's help"));
   
   dialog_setFontSize (SMALL);
   dialog_multiline (screen, screen->w, screen->h*1.25,
      _(
      "Use the the mouse's left button to place the blocks.\n"
      " \n"
      "The keys are:\n"
      "[0] Wall block\n"
      "[1] Ground block\n"
      "[2] Box block\n"
      "[3] Box right-placed block\n"
      "[4] End-point block\n"
      "[5] Place player (player can be placed only on a ground block)\n"
      "[Ctrl][s] Save level\n"
      )
                     );
   
   SDL_Flip (screen);
}

static void editor_help (SDL_Surface *screen)
{
   SDL_Event ev;
   
   editor_helpRender (screen); /* first draw to init display */
   
   
   while (SDL_WaitEvent (&ev))
   {
      switch (ev.type)
      {
         case SDL_QUIT: return;
         
         case SDL_KEYDOWN:
            switch (ev.key.keysym.sym)
            {
               case SDLK_q:
               case SDLK_ESCAPE: return;
               
               default:
                  continue;
                  break;
            }
            break;
         
         default:
            continue;
            break;
      }
      
      editor_helpRender (screen);
   }
}

static void editor_render (SDL_Surface *screen)
{
   map_blit (screen);
   map_blitObject (OBJECT (player_get ()), screen);
   //printf ("pp: %d, %d\n", player_getPosX (), player_getPosY ());
   
   /* trancparencify mouse's block */
   SDL_SetAlpha (_mouse->surface, SDL_SRCALPHA, 0x80 /*128*/);
   //_mouse->pos = _mousePos;
   object_setPosFromRect (_mouse, _mousePos);
   map_blitObject (_mouse, screen);
   /* and reset it's alpha to avoid "sepcial effect" :D */
   SDL_SetAlpha (_mouse->surface, SDL_SRCALPHA, 0xff /*255*/);
   
   //dialog_setFontSize (MEDIUM);
   //dialog_info (screen, screen->w, screen->h*1.5, "WIP/Not yet implemented");
   
   SDL_Flip (screen);
}


static MyBool editor_save (SDL_Surface *screen, char* defaultname)
{
   char* name;
   MyBool rv = TRUE;
   
   #ifdef DEBUG
   printf ("  Saving level...\n");
   #endif
   
   /* initialize font size for future dialogs */
   dialog_setFontSize (MEDIUM);
   
   /* ask user for level name */
   name = dialog_inputWithDefaultText (screen, defaultname,
                                       screen->w, screen->h);
   
   if (name != NULL)
   {
      char* filename;
      char* ext = "."LEVEL_EXT;
      
      filename = malloc (strlen (DATADIR) + strlen (name) + strlen (ext) + 1);
      
      if (filename != NULL)
      {
         if (sprintf (filename, "%s%s%s", DATADIR, name, ext) < 0)
         {
            free (filename);
            fprintf (stderr,
                     _("** Internal error: failed to create filename"));
            
            dialog_info (screen, screen->w, screen->h / 3,
                         _("Failed to save level"));
            
            rv = FALSE;
         }
         else
         {
            /* FIXME:
             * add created level to a level set */
            /* FIXME:
             * check if file already exists */
            
            if (level_write (filename, player_getPosX (), player_getPosY ()))
            {
               dialog_info (screen, screen->w, screen->h/3, _("Level saved"));
               
               /* try to add level to a set */
               //if (set_open (DATADIR"index.set") == 0)
               //{
                  char *levelname;
                  
                  levelname = malloc (strlen (name) + strlen (ext) + 1);
                  
                  if (levelname)
                  {
                     levelname = strcpy (levelname, name);
                     levelname = strcat (levelname, ext);
                     
                     if (!set_levelExists (levelname))
                     {
                        if (!set_appendLevel (levelname))
                           dialog_info (screen, screen->w, screen->h/2,
                                        _("Failed to add level to set"));
                     }
                     #ifdef DEBUG
                     else
                        printf ("level alreay in set\n");
                     #endif
                     
                     if (!set_write (DATADIR"index.set"))
                        printf ("failed fo writ set\n");
                     
                     //set_close ();
                     
                     free (levelname);
                  }
                  else
                  {
                     memalloc_errormsg ();
                     
                     rv = FALSE;
                  }
               //}
            }
            else
            {
               dialog_info (screen, screen->w, screen->h/3,
                            _("Failed to save level"));
               
               rv = FALSE;
            }
            
            free (filename);
         }
      }
      else
      {
         memalloc_errormsg ();
         rv = FALSE;
      }
      
      free (name);
      
      SDL_Flip (screen);
      
      SDL_Delay (2000);
   }
   else
      rv = FALSE;
   
   return rv;
}


void editor_main (SDL_Surface *screen)
{
   SDL_Event ev;
   int run = 1;
   Map block = GROUND; /* default block */
   
   if (set_open (DATADIR"index.set") != 0)
      return;
   
   player_init ();
   _mouse = gallery_getItem (block);
   
   editor_playerRemove (); /* hide the player by default */
   
   /* temporary fake to load user level */
   if (set_loadLastLevel () != 0)
      map_resetToBlock (WALL);
   
   editor_render (screen); /* init display */
   
   
   while (SDL_WaitEvent (&ev) && run)
   {
      switch (ev.type)
      {
         case SDL_QUIT:
            run = 0;
            continue;
            break;
         
         case SDL_KEYDOWN:
            switch (ev.key.keysym.sym)
            {
               case SDLK_q:
               case SDLK_ESCAPE:
                  run = 0;
                  continue;
                  break;
               
               case SDLK_F1:
               case SDLK_h:
                  editor_help (screen);
                  break;
               
               /* FIXME:
                * clean this code and put it in a function */
               case SDLK_s: /* save current level */
                  if (KMOD_CTRL & SDL_GetModState()) // if CTRL is active
                  {
                     //if (KMOD_SHIFT & SDL_GetModState())
                     //else {
                     
                     char* prev;
                     
                     /* try to get last level name */
                     if ((prev = set_getLastLevel ()))
                     {
                        /* on vire l'extension */
                        prev[strlen (prev) - strlen (LEVEL_EXT) - 1] = '\0';
                        
                        editor_save (screen, prev);
                        free (prev);
                     }
                     else
                        editor_save (screen, NULL);
                     
                     //}
                  }
                  break;
               
               case SDLK_KP0:
               case SDLK_0:
                  block = WALL;
                  _mouse = gallery_getItem (block);
                  break;
               
               case SDLK_KP1:
               case SDLK_1:
                  block = GROUND;
                  _mouse = gallery_getItem (block);
                  break;
               
               case SDLK_KP2:
               case SDLK_2:
                  block = BOX;
                  _mouse = gallery_getItem (block);
                  break;
               
               case SDLK_KP3:
               case SDLK_3:
                  block = BOXDONE;
                  _mouse = gallery_getItem (block);
                  break;
               
               case SDLK_KP4:
               case SDLK_4:
                  block = POINT;
                  _mouse = gallery_getItem (block);
                  break;
               
               case SDLK_KP5:
               case SDLK_5:
                  block = UNKNOWN;
                  /* le player qui suis la souris fait déconnet l'autre
                   * car ils utilisent le même objet, donc qd on change la pos
                   * de l'un, on le fait aussi sur l'autre.
                   * Faudra trouver une soluce */
                  //_mouse = OBJECT (player_get ());
                  break;
               
               case SDLK_r: /* reset map to WALL */
                  block = GROUND;
                  _mouse = gallery_getItem (block);
                  
                  editor_playerRemove ();
                  
                  map_resetToBlock (WALL);
                  break;
               
               default:
                  continue;
                  break;
            }
            break;
         
         case SDL_MOUSEBUTTONDOWN:
            switch (ev.button.button)
            {
               case SDL_BUTTON_LEFT:
                  /* place blocks */
                  if (block == UNKNOWN) // player
                  {
                     int x = ev.button.x / MAP_BLOCKSIZE;
                     int y = ev.button.y / MAP_BLOCKSIZE;
                     
                     if (map_checkBlockAtPos (x, y, GROUND | POINT))
                        player_setPos (x, y);
                  }
                  else
                     map_setBlockAtPos (ev.button.x / MAP_BLOCKSIZE,
                                        ev.button.y / MAP_BLOCKSIZE,
                                        block);
                  
                  break;
               /*
               case SDL_BUTTON_RIGHT:
                  // on efface ?
                  break;
               */
               default:
                  continue;
                  break;
            }
            break;
         
         case SDL_MOUSEMOTION:
            //editor_setMousePos (ev.motion.x + 10, ev.motion.y + 10);
            editor_setMousePos (ev.motion.x / MAP_BLOCKSIZE,
                                ev.motion.y / MAP_BLOCKSIZE);
            break;
         
         default:
            continue;
            break;
      }
      /* disactivate mouse events during render
       * FIXME check if this make click hard to get */
      SDL_EventState (SDL_MOUSEMOTION, SDL_DISABLE);
      
      editor_render (screen);
      
      /* reactivate mouse events */
      SDL_EventState (SDL_MOUSEMOTION, SDL_ENABLE);
   }
   
   
   _mouse = NULL;
   
   player_free ();
   
   set_close ();
}
